// Concurrent Web Proxy Tester
//
// Test both numeric and symbolic server names.
// Are they confused by ordinary http headers?
// Demand http/1.0? or plain "GET /file\n"?
//   Is 1.0 the right answer? Hopefully doesn't default to persistent.
// However, if we send proxy an http/1.1 request,
//   it should return an http/1.1 response.
// Demand error status messages for ordinary errors?
//
// Phase Two tests:
// Can 2nd xfer complete if first is hanging:
//   In request read.
//   In connect().
//   In response read.

#undef HAVE_CONFIG_H
#include "stdio.h"
#include "stdlib.h"
#include "amisc.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <string.h>
#include "async.h"
#include "rxx.h"

// Poor man's integer random number generator.
// The point is to be able to generate reproducible
// sequences without using srandom().
// Constants from Numerical Recipes in C.
class RNG {
public:
  RNG::RNG(int seed) { jran = seed; }
  int next() { jran = (jran * 3877 + 29573) % 139968; return(jran>>3); }
private:
  int jran;
};

class reply_thunk { public:
  int s;
  int servno;
  int index;
  int size;
  RNG *r;
  bool blocked;

  reply_thunk() { blocked = false; }
};

class writereq_thunk { public:
  int flags;
  int s;
  int servno;
  callback<void, bool>::ref cb;
  str data;
  char *buf;

  writereq_thunk(callback<void, bool>::ref c) : cb(c) {};
};


bool doinfinity = true;

void accept_cb(int s, callback<void, int>::ref cb);
void child_cb(char *progname, int status);
void done(bool ok);
void ws_cb(int s, int servno, rxx *r);
void write_req(int flags, int servno, callback<void, bool>::ref cb, int s);
void write_req_part1(writereq_thunk *wrx);
void write_req_part2(writereq_thunk *wrx);
void write_big_req(writereq_thunk *wrx, ssize_t written);
void start_inf_req(int flags, int s, callback<void, bool>::ref cb);
void write_inf_req(int flags, int s, callback<void, bool>::ref cb);
void write_inf_req_cb(int flags, int s, callback<void, bool>::ref cb);
void start_header_fetch(writereq_thunk *wrx);
void premature_close(int s, callback<void, bool>::ref cb);
void fetch_header_byte(writereq_thunk *wrx, int);
void process_header(writereq_thunk *wrx, int n, char *xb, int xn);
void start_check_body(writereq_thunk *wrx);
void check_body(writereq_thunk *wrx, RNG *r, int index, int size, char *buf,
    int len);

void ws_reply(int s, int flags, str file, str version);
void ws_reply_cb1(reply_thunk *);
void done_split_test(bool ok);
void infinite_done(bool ok);
void stress_test_iter(int, int, int, bool, bool);

#define TEST_ARGS bool ok
void fetch_by_name(TEST_ARGS);
void split_test(TEST_ARGS);
void large_req(TEST_ARGS);
void large_resp(TEST_ARGS);
void zero_test(TEST_ARGS);
void bad_port_test(TEST_ARGS);
void bad_port_test2(TEST_ARGS);
void malformed_request(TEST_ARGS);
void malformed_request_cb(TEST_ARGS);
void close_request(TEST_ARGS);
void close_request_cb(TEST_ARGS);
void infinite_request(TEST_ARGS);
void blocking_server(TEST_ARGS);
void blocking_server_cb(int *nreturns, int *oks);
void silent_client(int seqno, int *nreturns, int *oks, bool ok);
void silent_client_cb(int *nreturns, int *oks);
void large_and_small_request(int seqno, int *nreturns, int *oks, TEST_ARGS);
void large_and_small_request_cb(int *nreturns, timecb_t *to);
void do_nothing(TEST_ARGS);
void bad_connect_and_normal(int seqno, int *nreturns, int *oks, bool ok);
void bad_connect_and_normal_cb(timecb_t *to, int *nreturns);
void stress_test(int seqno, timecb_t *to, int *nreturns, TEST_ARGS);

int child_port;

// The client side of the tester indicates the file name it is fetching,
// the desired file length, and the seed to pass to RNG() to
// generated the content.


#define MAX_SERVERS 8
#define NORMAL_SERVER 0
#define NORMAL_SERVER2 1
#define NORMAL_SERVER3 2
#define BLOCKING_SERVER 3
#define INFINITE_SERVER 4
#define INFINITE_SERVER2 5
#define DEAD_SERVER 6
int servers[MAX_SERVERS];
int seeds[MAX_SERVERS];
int lengths[MAX_SERVERS];
str files[MAX_SERVERS];
int messages[MAX_SERVERS];

#define FETCH_SPLIT 0x1
#define FETCH_LARGE_REQ 0x2
#define FETCH_LARGE_RESP 0x4
#define FETCH_ZERO  0x8
#define FETCH_BAD_PORT  0x10
#define FETCH_INFINITE  0x20
#define FETCH_MALFORMED  0x40
#define FETCH_CLOSE 0x80
#define FETCH_INFINITE_RESP  0x100
#define FETCH_NAME  0x200
#define FETCH_DEAD_SERVER  0x400
#define FETCH_SILENT_CLIENT  0x800

#define BIG_REQ_LEN      (0xffff)
#define BIG_RESP_LEN     (20000000 + (random() % 1000000))
#define NORMAL_PAGE_LEN  (4096 + (random() % 20000))

// CHUNKSIZE has to be an order of magnitude smaller than NORMAL_PAGE_LEN so
// that BLOCKING_SERVER can block roughly halfway through the reply
#define CHUNKSIZE 1024

/*
 * Create a TCP server. Returns port the server listens to.
 * cb will be called for each newly accepted connection.
 */
void
make_server(int servno, callback<void, int>::ref cb)
{
  int s;
  unsigned sinlen;
  struct sockaddr_in sin;

  if(servno == DEAD_SERVER) {
    servers[servno] = 81;
    return;
  }

  s = inetsocket(SOCK_STREAM, 0, INADDR_ANY);
  assert(s >= 0);
  make_async(s);
  tcp_nodelay(s);
  listen(s, 128);

  sinlen = sizeof(sin);
  assert(getsockname(s, (struct sockaddr *) &sin, &sinlen) == 0);

  messages[servno] = 0;
  fdcb(s, selread, wrap(accept_cb, s, cb));

  servers[servno] = ntohs(sin.sin_port);
}

void
accept_cb(int s, callback<void, int>::ref cb)
{
  int s1;
  unsigned sinlen;
  struct sockaddr_in sin;

  sinlen = sizeof(sin);
  s1 = accept(s, (struct sockaddr *) &sin, &sinlen);
  assert(s1 >= 0);
  make_async(s1);
  tcp_nodelay(s1);

  cb(s1);
}

void
write_n_cb(int s, char *buf, int off, int len, callback<void>::ref cb)
{
  int cc;

  cc = write(s, buf + off, len);
  if(cc >= len){
    fdcb(s, selwrite, 0);
    free(buf);
    cb();
  } else if(cc > 0){
    fdcb(s, selwrite, wrap(write_n_cb, s, buf, off + cc, len - cc, cb));
  } else {
    fdcb(s, selwrite, 0);
    free(buf);
    cb();
  }
}

void
write_n(int s, const char *buf, int n, callback<void>::ref cb)
{
  char *p = (char *) malloc(n == 0 ? 1 : n);
  assert(p);
  bcopy(buf, p, n);
  fdcb(s, selwrite, wrap(write_n_cb, s, p, 0, n, cb));
}

void
read_n_cb(int s, char *p, int off, int n, callback<void, char *, int>::ref cb)
{
  int cc;

  fdcb(s, selread, 0);

  cc = read(s, p + off, n - off);
  if(cc <= 0 || off + cc >= n){
    cb(p, off + cc);
  } else {
    fdcb(s, selread, wrap(read_n_cb, s, p, off + cc, n, cb));
  }
}

void
read_n(int s, int n, callback<void, char *, int>::ref cb)
{
  char *p = (char *) malloc(n);
  assert(p);
  fdcb(s, selread, wrap(read_n_cb, s, p, 0, n, cb));
}

struct junk {
  int s;
  rxx *r;
  char *buf;
  int bufmax;
  int next;
  callback<void>::ref cb;
  junk(int xs, rxx *xr, char *xbuf, int xbufmax, int xnext,
       callback<void>::ref xcb) : cb(xcb) {
    s = xs; r = xr; buf = xbuf; bufmax = xbufmax;
    next = xnext;
  }
};

void
read_rxx_iter(junk *j)
{
  int cc = 0, n;

  n = j->bufmax - j->next;
  if(n > 0){
    cc = read(j->s, j->buf + j->next, n);
    if(cc > 0)
      j->next += cc;
  }

  j->r->clear();
  if(j->r->search(str(j->buf, j->next))){
    free(j->buf);
    fdcb(j->s, selread, 0);
    j->cb();
    delete j;
    return;
  }

  if(j->next >= j->bufmax || (cc < 0 && errno != EWOULDBLOCK)){
    fdcb(j->s, selread, 0);
    j->cb();
    free(j->buf);
    delete j;
  } else {
    fdcb(j->s, selread, wrap(read_rxx_iter, j));
  }
}

void
read_rxx(int s, rxx *r, callback<void>::ref cb)
{
  int bufmax = 4096; // XXX
  char *buf = (char *) malloc(bufmax);
  assert(buf);
  read_rxx_iter(new junk(s, r, buf, bufmax, 0, cb));
}

// The web proxy has connected to our web server;
// read the request. Be prepared to parse either an
// old-style or a new-style request.
void
web_server(int servno, int s)
{
  rxx *r = new rxx("^(get\\s+(\\S+)[\\r\\n])|(get\\s+(\\S+)\\s+(http/\\S+))", "i");
  read_rxx(s, r, wrap(ws_cb, s, servno, r));
}

// The web server has read the first line of a request from the proxy
// (at least that's what the regular expression thinks).
void
ws_cb(int s, int servno, rxx *r)
{
  if(r->len(1) > 0){
    // old-style GET (no HTTP/1.0, no headers)
    // so response must not contain headers.
    ws_reply(s, servno, (*r)[2], "");
  } else if(r->len(3) > 0){
    // new-style GET with HTTP/x.y and headers.
    // so response must contain status and headers.
    ws_reply(s, servno, (*r)[4], (*r)[5]);
  } else {
    warn << "wpt1 server: cannot parse request\n";
    close(s);
  }

  delete r;
}

// Send a reply back to the proxy; the proxy asked for file,
// and it sent the GET with the indicated HTTP version,
// e.g. "HTTP/1.0".
void
ws_reply(int s, int servno, str file, str version)
{
  if(file != files[servno]){
    warn << "wpt1: file " << file << " wanted_file " << files[servno] << "\n";
    close(s);
    return;
  }

  char p[512];
  if(version.len() > 0){
    sprintf(p, "%s 200 OK\r\n\r\n", version.cstr());
  } else {
    p[0] = '\0';
  }

  reply_thunk *rtx = New reply_thunk();
  rtx->s = s;
  rtx->servno = servno;
  rtx->index = 0;
  rtx->size = lengths[servno];
  rtx->r = new RNG(seeds[servno]);
  write_n(s, p, strlen(p), wrap(ws_reply_cb1, rtx));
}

// Write a chunk of the reply body to the proxy.
// Index indicates the byte number of the next byte to write.
// Size indicates the total number of bytes in the body.
// The body bytes are randomly generated in a way that
// the client can check them; the RNG supplies the
// stream of bytes.
void
ws_reply_cb1(reply_thunk *rtx)
{
  char buf[CHUNKSIZE];
  int i;

  // infinite server keeps writing CHUNKSIZE bytes at a time until
  // messages[INFINITE_SERVER] has a value
  if(rtx->servno == INFINITE_SERVER || rtx->servno == INFINITE_SERVER2) {
    if(messages[rtx->servno]) {
      close(rtx->s);
      delete rtx->r;
      messages[rtx->servno] = 0;
      delete rtx;
      return;
    }

    for(i = 0; i < CHUNKSIZE; i++)
      buf[i] = rtx->r->next();

  // blocking server is almost a normal server, but blocks halfway through the
  // reply for 15 seconds
  } else if(rtx->servno == BLOCKING_SERVER &&
      (rtx->index >= rtx->size/2) &&
      !rtx->blocked)
  {
    rtx->blocked = true;
    delaycb(15, 0, wrap(ws_reply_cb1, rtx));
    return;

  // blocking and normal server have to stop when done
  } else if(rtx->index >= rtx->size) {
    close(rtx->s);
    delete rtx->r;
    delete rtx;
    return;

  // normal server.
  } else {
    for(i = 0; i < CHUNKSIZE && rtx->index < rtx->size; i++, rtx->index++)
      buf[i] = rtx->r->next();
  }

  write_n(rtx->s, buf, i, wrap(ws_reply_cb1, rtx));
}

int child_pid;

void
start_child(char *progname)
{
  int pid;
  char tbuf[64];

  sprintf(tbuf, "%d", child_port);

  pid = fork();
  if(pid < 0){
    perror("tprox: fork");
    exit(1);
  }
  if(pid == 0){
    struct rlimit rl;
    int i;
    for(i = 3; i < 100; i++)
      close(i);
#if 0
    // suppress output from proxy
    int fd = open("/dev/null", 2);
    dup2(fd, 0);
    dup2(fd, 1);
    dup2(fd, 2);
#endif
    getrlimit(RLIMIT_DATA, &rl);
    if(rl.rlim_cur > 100 * 1024 * 1024)
      rl.rlim_cur = 100 * 1024 * 1024;
    rl.rlim_max = 100 * 1024 * 1024;
    if(setrlimit(RLIMIT_DATA, &rl) < 0)
      perror("setrlimit");

    rl.rlim_cur = 1;
    rl.rlim_max = 1;
    if(setrlimit(RLIMIT_NPROC, &rl) < 0)
      perror("setrlimit");

    execl(progname, progname, tbuf, 0);
    perror("tprox: could not exec tcpproxy program");
    exit(1);
  }
  child_pid = pid;
  sleep(2);
  chldcb(pid, wrap(child_cb, progname));
}

void
child_cb(char *progname, int status)
{
  warn << "wpt1: child webproxy failed; cannot continue.\n";
  done(false);
}

void
proxy_connect(int flags, int servno, callback<void, bool>::ref cb)
{
  if(flags & FETCH_ZERO)
    lengths[servno] = 0;
  else if(flags & FETCH_LARGE_RESP)
    lengths[servno] = BIG_RESP_LEN;
  else if(flags & FETCH_INFINITE || flags & FETCH_INFINITE_RESP)
    lengths[servno] = 0x7fffffff;
  else
    lengths[servno] = NORMAL_PAGE_LEN;
  
  char fb[512];
  sprintf(fb, "/%d", (int) random());
  files[servno] = fb;
  seeds[servno] = random();

  if(flags & FETCH_NAME) {
    tcpconnect("localhost", child_port, wrap(write_req, flags, servno, cb));
  } else {
    tcpconnect("127.0.0.1", child_port, wrap(write_req, flags, servno, cb));
  }
}

int randport()
{
  int port;
  while(1) {
    port = random();
    if(port <= 1024 || port >= 65535)
      continue;

    for(int i=0; i<MAX_SERVERS; i++)
      if(servers[i] == port)
        continue;

    break;
  };
  
  return port;
}

void
write_req(int flags, int servno, callback<void, bool>::ref cb, int s)
{
  char urlbuf[4096];

  if(s < 0) {
    perror("wpt1: write_req: connect to proxy failed");
    cb(false);
    return;
  }

  if((flags & FETCH_INFINITE) || (flags & FETCH_LARGE_REQ)) {
    // missing \r\n
    sprintf(urlbuf, "GET http://%s:%d%s HTTP/1.0\r\n",
        (flags & FETCH_DEAD_SERVER) ? "213.84.252.212" : "127.0.0.1",
        servers[servno], (const char *) files[servno]);
  } else if(flags & FETCH_MALFORMED){
    unsigned int i;
    for(i = 0; i < 30; i++)
      urlbuf[i] = 'a' + (random() % 26);
    urlbuf[i++] = '\r';
    urlbuf[i++] = '\n';
    urlbuf[i++] = '\r';
    urlbuf[i++] = '\n';
    urlbuf[i] = '\0';
  } else {
    sprintf(urlbuf, "GET http://%s:%d%s HTTP/1.0\r\n\r\n",
        (flags & FETCH_DEAD_SERVER) ? "213.84.252.212" : "127.0.0.1",
            (flags & FETCH_BAD_PORT) ? randport() : servers[servno],
            (const char *) files[servno]);
  }


  if(flags & FETCH_SPLIT || flags & FETCH_SILENT_CLIENT){
    int ulen = strlen(urlbuf);
    writereq_thunk *wrx = new writereq_thunk(cb);
    assert(wrx);
    wrx->flags = flags;
    wrx->s = s;
    wrx->data = str(urlbuf + ulen - 15);
    wrx->servno = servno;
    write_n(s, urlbuf, ulen - 15, wrap(write_req_part1, wrx));

  } else if(flags & FETCH_LARGE_REQ){
    writereq_thunk *wrx = new writereq_thunk(cb);
    assert(wrx);
    wrx->flags = flags;
    wrx->s = s;
    wrx->data = "";
    wrx->servno = servno;

    write_n(s, urlbuf, strlen(urlbuf),
            wrap(write_big_req, wrx, strlen(urlbuf)));

  } else if(flags & FETCH_INFINITE){
    write_n(s, urlbuf, strlen(urlbuf),
            wrap(start_inf_req, flags, s, cb));

  } else if(flags & FETCH_CLOSE){
    write_n(s, urlbuf, random() % (strlen(urlbuf) + 1),
            wrap(premature_close, s, cb));

  } else {
    writereq_thunk *wrx = new writereq_thunk(cb);
    assert(wrx);
    wrx->flags = 0;
    wrx->s = s;
    wrx->data = "";
    wrx->servno = servno;
    write_n(s, urlbuf, strlen(urlbuf), wrap(start_header_fetch, wrx));
  }
}

// As part of a malformed request, just close the connection now.
void
premature_close(int s, callback<void, bool>::ref cb)
{
  close(s);
  cb(false);
}

// Chunk of request has been sent. Delay before sending chunk 2.
void
write_req_part1(writereq_thunk *wrx)
{
  int delay = (wrx->flags & FETCH_SILENT_CLIENT) ? 15 : 1;
  delaycb(delay, 0, wrap(write_req_part2, wrx));
}

// Send second chunk of request
void
write_req_part2(writereq_thunk *wrx)
{
  write_n(wrx->s, wrx->data.cstr(), wrx->data.len(), wrap(start_header_fetch,
        wrx));
}

// Write big request
void
write_big_req(writereq_thunk *wrx, ssize_t written)
{
  char bsbuf[BIG_REQ_LEN - written];
  unsigned i = 0;
  
  bsbuf[i++] = 'A' + (random() % 26);
  for(; i < (unsigned) (random() % 16); i++)
    bsbuf[i] = 'a' + (random() % 26);
  bsbuf[i++] = ':';
  bsbuf[i++] = ' ';
  for(; i < sizeof(bsbuf)-4; i++)
    bsbuf[i] = 'a' + (random() % 26);
  bsbuf[i++] = '\r';
  bsbuf[i++] = '\n';
  bsbuf[i++] = '\r';
  bsbuf[i++] = '\n';

  write_n(wrx->s, bsbuf, sizeof(bsbuf), wrap(start_header_fetch, wrx));
}

// Write endless junk to the proxy as part of a request.
// But stop after 10 seconds.
void
start_inf_req(int flags, int s, callback<void, bool>::ref cb)
{
  // 10 seconds should be enough to swamp the sucker
  delaycb(10, 0, wrap(write_inf_req_cb, flags, s, cb));
  write_inf_req(flags, s, cb);
}

void
write_inf_req(int flags, int s, callback<void, bool>::ref cb)
{
  char bsbuf[CHUNKSIZE];
  unsigned i = 0;

  bsbuf[i++] = 'A' + (random() % 26);
  for(; i < (unsigned) (random() % 16); i++)
    bsbuf[i] = 'a' + (random() % 26);
  bsbuf[i++] = ':';
  bsbuf[i++] = ' ';
  for(; i < sizeof(bsbuf)-3; i++)
    bsbuf[i] = 'a' + (random() % 26);
  bsbuf[i++] = '\r';
  bsbuf[i++] = '\n';
  bsbuf[i++] = '\0';

  write_n(s, bsbuf, sizeof(bsbuf), wrap(write_inf_req, flags, s, cb));
}

// does proxy survive large request after infinite request?
void
write_inf_req_cb(int flags, int s, callback<void, bool>::ref cb)
{
  fdcb(s, selwrite, 0);
  close(s);
  proxy_connect(FETCH_LARGE_RESP, NORMAL_SERVER, cb);
}

// Start reading the response header from the proxy.
// Read it one byte at a time to avoid consuming any
// of the body. All we're really doing is checking
// for \r\n.
void
start_header_fetch(writereq_thunk *wrx)
{
  wrx->buf = (char *) malloc(CHUNKSIZE);
  assert(wrx->buf);
  fetch_header_byte(wrx, 0);
}

// We've read n bytes of the response header from the proxy
// into buf[].
void
// fetch_header_byte(int s, int servno, callback<void, bool>::ref cb, char *buf, int n)
fetch_header_byte(writereq_thunk *wrx, int n)
{
  if(n >= CHUNKSIZE){
    free(wrx->buf);
    close(wrx->s);
    wrx->cb(false);
    delete wrx;
    return;
  }

  read_n(wrx->s, 1, wrap(process_header, wrx, n));
}

// read_n() just read one byte from the response header
// into xb[0]. All we're trying to do is stop as soon as
// we see \r\n\r\n.
void
process_header(writereq_thunk *wrx, int n, char *xb, int xn)
{
  if(xn < 1) {
    free(wrx->buf);
    close(wrx->s);
    wrx->cb(false);
    delete wrx;
  } else {
    wrx->buf[n++] = xb[0];
    if(n >= 4 && memcmp(wrx->buf + n - 4, "\r\n\r\n", 4) == 0){
      free(wrx->buf);
      start_check_body(wrx);
    } else {
      fetch_header_byte(wrx, n);
    }
  }

  free(xb);
}

// We've read the response header from the proxy.
// Read the body and make sure
// it's the same pseudo-random data stream that
// the server generated.
void
start_check_body(writereq_thunk *wrx)
{
  read_n(wrx->s, CHUNKSIZE, wrap(check_body, wrx, new RNG(seeds[wrx->servno]),
        0, lengths[wrx->servno]));
}

void
check_body(writereq_thunk *wrx, RNG *r, int index, int size, char *buf, int len)
{
  int i, bad = 0;

  // if infinite request, then don't bother to check. just keep sucking it in.
  if(wrx->flags & FETCH_INFINITE_RESP) {
    if(buf)
      free(buf);
    read_n(wrx->s, CHUNKSIZE, wrap(check_body, wrx, r, len, size));
    return;
  }

  for(i = 0; i < len; i++)
    if((buf[i] & 0xff) != (r->next() & 0xff))
      bad = 1;

  if(buf)
    free(buf);

  if(index + len == size && bad == 0){
    close(wrx->s);
    delete r;
    wrx->cb(true);
    delete wrx;
  } else if(bad || len < 1 || index + len > size){
    close(wrx->s);
    delete r;
    wrx->cb(false);
    delete wrx;
  } else {
    read_n(wrx->s, CHUNKSIZE, wrap(check_body, wrx, r, index + len, size));
  }
}


void
get_test()
{
  printf("Ordinary fetch: ");
  proxy_connect(0, NORMAL_SERVER, wrap(fetch_by_name));
}

void
fetch_by_name(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");

  printf("Fetch by name: ");
  proxy_connect(FETCH_NAME, NORMAL_SERVER, wrap(split_test));
}


void
split_test(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");

  printf("Split request: ");
  proxy_connect(FETCH_SPLIT, NORMAL_SERVER, wrap(large_req));
}


void
large_req(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");

  printf("Large request: ");
  proxy_connect(FETCH_LARGE_REQ, NORMAL_SERVER, wrap(large_resp));
}


void
large_resp(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");

  printf("Large response: ");
  proxy_connect(FETCH_LARGE_RESP, NORMAL_SERVER, wrap(zero_test));
}

void
zero_test(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");

  printf("Zero-size response: ");
  proxy_connect(FETCH_ZERO, NORMAL_SERVER, wrap(bad_port_test));
}

void
bad_port_test(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");

  printf("Recover after bad connect: ");
  proxy_connect(FETCH_BAD_PORT, NORMAL_SERVER, wrap(bad_port_test2));
}


void
bad_port_test2(bool ok)
{
  assert(!ok);
  proxy_connect(0, NORMAL_SERVER, wrap(malformed_request));
}

void
malformed_request(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");

  // shouldn't take longer than 10 seconds
  printf("Malformed request: ");
  proxy_connect(FETCH_MALFORMED, NORMAL_SERVER, wrap(malformed_request_cb));
}


// check if still OK after malformed request
void
malformed_request_cb(bool ok)
{
  proxy_connect(0, NORMAL_SERVER, wrap(close_request));
}

void
close_request(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");
  printf("Premature client close(): ");

  proxy_connect(FETCH_CLOSE, NORMAL_SERVER, wrap(close_request_cb));
}

void
close_request_cb(bool ok)
{
  proxy_connect(0, NORMAL_SERVER, wrap(infinite_request));
}

void
infinite_request(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");
  printf("Infinitely long request: ");

  proxy_connect(FETCH_INFINITE, NORMAL_SERVER, wrap(blocking_server));
}

void
do_it(callback<void, bool>::ref cb, bool ok)
{
  cb(ok);
}

void
delay_a_bit(callback<void, bool>::ref cb, bool ok)
{
  delaycb(3, 0, wrap(do_it, cb, ok));
}


void
blocking_server(bool ok)
{
  printf("%s\n", ok ? "pass" : "fail");
  printf("Blocking server, concurrent request: ");

  int *nreturns = new int;
  *nreturns = 0;

  int *oks = new int;
  *oks = 0;

  proxy_connect(0, BLOCKING_SERVER, wrap(silent_client, 1, nreturns, oks));
  delaycb(2, 0, wrap(blocking_server_cb, nreturns, oks));
}

void
blocking_server_cb(int *nreturns, int *oks)
{
  proxy_connect(0, NORMAL_SERVER, wrap(delay_a_bit, wrap(silent_client, 2,
          nreturns, oks)));
}

void
silent_client(int seqno, int *nreturns, int *oks, bool ok)
{
  // answer from normal server should come first
  if(seqno == 2 && !(*nreturns))
    (*oks)++;

  // answer from blocking server should come second
  if(seqno == 1 && (*nreturns) == 1)
    (*oks)++;

  if(++(*nreturns) < 2)
    return;

  printf("%s\n", (*oks == 2) ? "pass" : "fail");

  *oks = 0;
  *nreturns = 0;

  printf("Silent client, concurrent request: ");
  proxy_connect(FETCH_SILENT_CLIENT, NORMAL_SERVER,
      wrap(bad_connect_and_normal, 1, nreturns, oks));
  delaycb(2, 0, wrap(silent_client_cb, nreturns, oks));
}

void
silent_client_cb(int *nreturns, int *oks)
{
  proxy_connect(0, NORMAL_SERVER2, wrap(delay_a_bit,
        wrap(bad_connect_and_normal, 2, nreturns, oks)));
}


void
bad_connect_and_normal(int seqno, int *nreturns, int *oks, bool ok)
{
  // answer from second request should come first
  if(seqno == 2 && !(*nreturns))
    (*oks)++;

  // answer from first request should come second
  if(seqno == 1 && (*nreturns) == 1)
    (*oks)++;

  if(++(*nreturns) < 2)
    return;

  printf("%s\n", (*oks == 2) ? "pass" : "fail");


  int *nreturns2 = new int;
  *nreturns2 = 0;

  timecb_t *to2 = delaycb(40, 0,
      wrap(stress_test, 1, (timecb_t*) 0, nreturns2, false));

  printf("Unresponsive server, concurrent request: ");
  proxy_connect(FETCH_DEAD_SERVER, DEAD_SERVER,
      wrap(stress_test, 2, to2, nreturns2));
  delaycb(2, 0, wrap(bad_connect_and_normal_cb, to2, nreturns2));
}

void
bad_connect_and_normal_cb(timecb_t *to, int *nreturns)
{
  proxy_connect(0, NORMAL_SERVER, wrap(delay_a_bit, wrap(stress_test, 3, to,
          nreturns)));
}


// restart if they finish
void
inf_done(int servno, bool)
{
  proxy_connect(FETCH_INFINITE_RESP, servno, wrap(inf_done, servno));
}

void
stress_test(int seqno, timecb_t *to, int *nreturns, bool ok)
{
  if(++(*nreturns) > 1)
    return;

  printf("%s\n", (seqno == 3 && ok) ? "pass" : "fail");

  printf("Stress test: ");
  proxy_connect(FETCH_INFINITE_RESP, INFINITE_SERVER, wrap(inf_done, INFINITE_SERVER));
  proxy_connect(FETCH_INFINITE_RESP, INFINITE_SERVER2, wrap(inf_done, INFINITE_SERVER2));
  stress_test_iter(BLOCKING_SERVER, 0, 8, true, true);
  stress_test_iter(NORMAL_SERVER, 0, 256 + (random() % 128), true, true);
  stress_test_iter(NORMAL_SERVER2, 0, 256 + (random() % 128), true, true);
  stress_test_iter(NORMAL_SERVER3, 0, 256 + (random() % 128), true, true);
}


// Just finished test number "done" out of "wanted".
void
stress_test_iter(int servno, int n_done, int n_wanted, bool result_wanted,
    bool result)
{
  if(n_done >= n_wanted || result != result_wanted){
    messages[INFINITE_SERVER] = 1;
    messages[INFINITE_SERVER2] = 1;
    done((result == result_wanted));
  } else {
    if((n_done % 50) == 0)
      printf(".");
    if((n_done % 150) == 0) {
      proxy_connect(FETCH_LARGE_RESP, servno, wrap(stress_test_iter, servno,
                                           n_done + 1, n_wanted, true));
    } else if((n_done % 10) == 0) {
      proxy_connect(FETCH_MALFORMED, servno,
          wrap(stress_test_iter, servno, n_done + 1, n_wanted, false));

    } else if((n_done % 5) == 0) {
      proxy_connect(FETCH_SPLIT, servno,
          wrap(stress_test_iter, servno, n_done + 1, n_wanted, true));

    } else {
      proxy_connect(0, servno,
          wrap(stress_test_iter, servno, n_done + 1, n_wanted, true));
    }
  }
}



void
done(bool ok)
{
  static int n = 0;

  if(!ok) {
    printf("fail\n");
    goto done;
  }

  if(++n == 4)
    printf("pass\n");
  else
    return;

done:
  if(child_pid > 0)
    kill(child_pid, 9);
  exit(0);
}

int
main(int argc, char **argv)
{
  char *progname;

  if(argc != 2){
    fprintf(stderr, "Usage: tprox ./proxy\n");
    exit(1);
  }
  progname = argv[1];

  srandom(time(0) ^ (getpid() + (getpid() << 15)));

  setbuf(stdout, 0);
  child_port = 40000 + (random() % 20000);
  make_server(NORMAL_SERVER, wrap(web_server, NORMAL_SERVER));
  make_server(NORMAL_SERVER2, wrap(web_server, NORMAL_SERVER2));
  make_server(NORMAL_SERVER3, wrap(web_server, NORMAL_SERVER3));
  make_server(BLOCKING_SERVER, wrap(web_server, BLOCKING_SERVER));
  make_server(INFINITE_SERVER, wrap(web_server, INFINITE_SERVER));
  make_server(INFINITE_SERVER2, wrap(web_server, INFINITE_SERVER2));
  make_server(DEAD_SERVER, wrap(web_server, DEAD_SERVER));

  start_child(progname);

  get_test();

  amain();
}
